from ..base import BaseModel


class SampleSequence(BaseModel):
    stop_reason: str
    """停止原因"""

    tokens: list[int]
    """生成的 token id 列表"""

    text: str
    """生成的文本"""

    logprobs: list[float]
    """生成的 token 对应的 logprobs 列表"""
