import click


@click.group()
def cli():
    """Trio CLI"""
    pass


def register_cli() -> None:
    from .auth import login, logout

    cli.add_command(
        click.command(
            name="login",
        )(login)
    )
    cli.add_command(
        click.command(
            name="logout",
        )(logout)
    )


register_cli()
